@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2017 Fabian Groffen
+ * Copyright 2013-2022 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -33,10 +33,136 @@
 #ifdef HAVE_SSL
 #include <openssl/ssl.h>
 #include <openssl/err.h>
-#endif
-
 
 static char
+ssllisten(listener *lsnr)
+{
+	/* create a auto-negotiate context */
+	const SSL_METHOD *m;
+
+#if defined(HAVE_TLS_SERVER_METHOD)
+	m = TLS_server_method();
+#elif defined(HAVE_SSLV23_SERVER_METHOD)
+	m = SSLv23_server_method();
+#else
+# error "we don't understand this version of OpenSSL"
+#endif
+	lsnr->ctx = SSL_CTX_new(m);
+	if (lsnr->ctx == NULL) {
+		char *err = ERR_error_string(ERR_get_error(), NULL);
+		logerr("failed to create SSL context for listener on port %d: %s\n",
+				lsnr->port, err);
+		return 1;
+	}
+
+#if defined(HAVE_TLS_SERVER_METHOD)
+	/* handle possible restrictions to the flexible version method */
+	if (lsnr->protomin != 0) {
+		if (SSL_CTX_set_min_proto_version(lsnr->ctx, lsnr->protomin) == 0) {
+			char *err = ERR_error_string(ERR_get_error(), NULL);
+			logerr("SSL protomin version invalid: %s\n", err);
+			return 1;
+		}
+	}
+	if (lsnr->protomax != 0) {
+		if (SSL_CTX_set_max_proto_version(lsnr->ctx, lsnr->protomax) == 0) {
+			char *err = ERR_error_string(ERR_get_error(), NULL);
+			logerr("SSL protomax version invalid: %s\n", err);
+			return 1;
+		}
+	}
+#elif defined(HAVE_SSLV23_SERVER_METHOD)
+	/* handle restrictions by mapping the range of versions manually */
+	if (lsnr->protomin > 0 || lsnr->protomax > 0) {
+		long options = 0;
+
+		/* this is OpenSSL <1.1 which doesn't understand TLS1.3, we have
+		 * available:
+		 * SSL_OP_NO_SSLv2,
+		 * SSL_OP_NO_SSLv3,
+		 * SSL_OP_NO_TLSv1,
+		 * SSL_OP_NO_TLSv1_1 and
+		 * SSL_OP_NO_TLSv1_2 */
+		if (lsnr->protomax > 0) {
+			switch (lsnr->protomax) {
+				case SSL3_VERSION:
+					options |= SSL_OP_NO_TLSv1;
+				case TLS1_VERSION:
+					options |= SSL_OP_NO_TLSv1_1;
+				case TLS1_1_VERSION:
+					options |= SSL_OP_NO_TLSv1_2;
+				case TLS1_2_VERSION:
+				case TLS1_3_VERSION:
+					break;
+			}
+		}
+		if (lsnr->protomin > 0) {
+			switch (lsnr->protomin) {
+				case TLS1_3_VERSION:
+					options |= SSL_OP_NO_TLSv1_2;
+				case TLS1_2_VERSION:
+					options |= SSL_OP_NO_TLSv1_1;
+				case TLS1_1_VERSION:
+					options |= SSL_OP_NO_TLSv1;
+				case TLS1_VERSION:
+					options |= SSL_OP_NO_SSLv3;
+				case SSL3_VERSION:
+					break;
+			}
+		}
+		SSL_CTX_set_options(lsnr->ctx, options);
+	}
+#else
+# error "we don't understand this version of OpenSSL"
+#endif
+#ifdef HAVE_SSL_CTX_SET_CIPHER_LIST
+	if (lsnr->ciphers != NULL) {
+		if (SSL_CTX_set_cipher_list(lsnr->ctx, lsnr->ciphers) == 0) {
+			char *err = ERR_error_string(ERR_get_error(), NULL);
+			logerr("cannot set SSL cipher list: %s\n", err);
+			return 1;
+		}
+	}
+#endif
+#ifdef HAVE_SSL_CTX_SET_CIPHERSUITES
+	if (lsnr->ciphersuites != NULL) {
+		if (SSL_CTX_set_ciphersuites(lsnr->ctx, lsnr->ciphersuites) == 0) {
+			char *err = ERR_error_string(ERR_get_error(), NULL);
+			logerr("cannot set TLS ciphersuites: %s\n", err);
+			return 1;
+		}
+	}
+#endif
+
+	/* load certificates */
+	if (SSL_CTX_use_certificate_file(
+				lsnr->ctx, lsnr->pemcert, SSL_FILETYPE_PEM) <= 0)
+	{
+		char *err = ERR_error_string(ERR_get_error(), NULL);
+		SSL_CTX_free(lsnr->ctx);
+		logerr("cannot load certificate from %s: %s\n", lsnr->pemcert, err);
+		return 1;
+	}
+	if (SSL_CTX_use_PrivateKey_file(
+				lsnr->ctx, lsnr->pemcert, SSL_FILETYPE_PEM) <= 0)
+	{
+		char *err = ERR_error_string(ERR_get_error(), NULL);
+		SSL_CTX_free(lsnr->ctx);
+		logerr("cannot load private key from %s: %s\n", lsnr->pemcert, err);
+		return 1;
+	}
+	if (!SSL_CTX_check_private_key(lsnr->ctx)) {
+		char *err = ERR_error_string(ERR_get_error(), NULL);
+		SSL_CTX_free(lsnr->ctx);
+		logerr("private key does not match public certificate in %s: %s\n",
+				lsnr->pemcert, err);
+		return 1;
+	}
+	return 0;
+}
+#endif
+
+static int
 bindlistenip(listener *lsnr, unsigned int backlog)
 {
 	int sock;
@@ -48,41 +174,8 @@ bindlistenip(listener *lsnr, unsigned int backlog)
 	int sockcur = 0;
 
 #ifdef HAVE_SSL
-	if (lsnr->transport == W_SSL) {
-		/* create a auto-negotiate context */
-		const SSL_METHOD *m = SSLv23_server_method();
-		lsnr->ctx = SSL_CTX_new(m);
-		if (lsnr->ctx == NULL) {
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			logerr("failed to create SSL context for listener on port %d: %s\n",
-					lsnr->port, err);
-			return 1;
-		}
-
-		/* load certificates */
-		if (SSL_CTX_use_certificate_file(
-					lsnr->ctx, lsnr->pemcert, SSL_FILETYPE_PEM) <= 0)
-		{
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			SSL_CTX_free(lsnr->ctx);
-			logerr("cannot load certificate from %s: %s\n", lsnr->pemcert, err);
-			return 1;
-		}
-		if (SSL_CTX_use_PrivateKey_file(
-					lsnr->ctx, lsnr->pemcert, SSL_FILETYPE_PEM) <= 0)
-		{
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			SSL_CTX_free(lsnr->ctx);
-			logerr("cannot load private key from %s: %s\n", lsnr->pemcert, err);
-			return 1;
-		}
-		if (!SSL_CTX_check_private_key(lsnr->ctx)) {
-			char *err = ERR_error_string(ERR_get_error(), NULL);
-			SSL_CTX_free(lsnr->ctx);
-			logerr("private key does not match public certificate in %s: %s\n",
-					lsnr->pemcert, err);
-			return 1;
-		}
+	if ((lsnr->transport & ~0xFFFF) == W_SSL && ssllisten(lsnr)) {
+		return 1;
 	}
 #endif
 
@@ -165,12 +258,18 @@ bindlistenip(listener *lsnr, unsigned int backlog)
 	return 0;
 }
 
-static char
+static int
 bindlistenunix(listener *lsnr, unsigned int backlog)
 {
 	struct sockaddr_un srvr;
 	int sock;
 
+#ifdef HAVE_SSL
+	if ((lsnr->transport & ~0xFFFF) == W_SSL && ssllisten(lsnr)) {
+		return 1;
+	}
+#endif
+
 #ifndef PF_LOCAL
 # define PF_LOCAL PF_UNIX
 #endif
@@ -210,7 +309,7 @@ bindlistenunix(listener *lsnr, unsigned int backlog)
  * Open up sockets associated with listener.  Returns 0 when opening up
  * the listener succeeded, 1 otherwise.
  */
-char
+int
 bindlisten(listener *lsnr, unsigned int backlog)
 {
 	switch (lsnr->ctype) {
@@ -231,6 +330,10 @@ close_socks(listener *lsnr)
 	int i;
 	for (i = 0; lsnr->socks[i] != -1; i++)
 		close(lsnr->socks[i]);
+#ifdef HAVE_SSL
+	if ((lsnr->transport & ~0xFFFF) == W_SSL)
+		SSL_CTX_free(lsnr->ctx);
+#endif
 	logout("closed listener for %s %s:%u\n",
 			lsnr->ctype == CON_UDP ? "udp" : "tcp",
 			lsnr->ip == NULL ? "" : lsnr->ip, lsnr->port);
@@ -241,6 +344,10 @@ destroy_usock(listener *lsnr)
 {
 	close(lsnr->socks[0]);
 	unlink(lsnr->ip);
+#ifdef HAVE_SSL
+	if ((lsnr->transport & ~0xFFFF) == W_SSL)
+		SSL_CTX_free(lsnr->ctx);
+#endif
 	logout("removed listener for %s\n", lsnr->ip);
 }
 
