@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2017 Fabian Groffen
+ * Copyright 2013-2022 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -25,6 +25,7 @@
 #include <signal.h>
 #include <time.h>
 #include <sys/socket.h>
+#include <sys/stat.h>
 #include <errno.h>
 #include <assert.h>
 
@@ -41,11 +42,17 @@
 #include "aggregator.h"
 #include "collector.h"
 #include "conffile.h"
+#include "posixregex.h"
 
 unsigned char keep_running = 1;
 int pending_signal = -1;
 char relay_hostname[256];
 unsigned char mode = 0;
+char noexpire = 0;
+#ifdef HAVE_SSL
+char *sslCA = NULL;
+char sslCAisdir = 0;
+#endif
 
 static char *config = NULL;
 static int batchsize = 2500;
@@ -67,6 +74,9 @@ static FILE *relay_stdout = NULL;
 static FILE *relay_stderr = NULL;
 static char relay_can_log = 0;
 
+#ifdef YYDEBUG
+extern int router_yydebug;
+#endif
 
 /**
  * Writes to the setup output stream, prefixed with a timestamp, and if
@@ -110,7 +120,8 @@ relaylog(enum logdst dest, const char *fmt, ...)
 	len = strftime(prefix, sizeof(prefix), "[%Y-%m-%d %H:%M:%S]", &tm_now);
 
 	if (!console)
-		(void)snprintf(prefix + len, sizeof(prefix) - len, " (%s)", dest == LOGOUT ? "MSG" : "ERR");
+		(void)snprintf(prefix + len, sizeof(prefix) - len,
+				" (%s)", dest == LOGOUT ? "MSG" : "ERR");
 
 	fprintf(dst, "%s ", prefix);
 
@@ -187,7 +198,6 @@ do_reload(void)
 	for ( ; lsnrs != NULL; lsnrs = lsnrs->next) {
 		if (router_contains_listener(newrtr, lsnrs) == NULL) {
 			dispatch_removelistener(lsnrs);
-			shutdownclose(lsnrs);
 		}
 	}
 
@@ -300,6 +310,9 @@ handle_signal(int sig)
 		case SIGHUP:
 			sign = "SIGHUP";
 			break;
+		case SIGUSR1:
+			sign = "SIGUSR1";
+			break;
 	}
 	if (keep_running) {
 		logout("caught %s\n", sign);
@@ -311,6 +324,12 @@ handle_signal(int sig)
 	}
 	if (sig == SIGHUP) {
 		do_reload();
+	} else if (sig == SIGUSR1) {
+		if (mode & MODE_DEBUG) {
+			mode &= ~MODE_DEBUG;
+		} else {
+			mode |= MODE_DEBUG;
+		}
 	} else {
 		keep_running = 0;
 	}
@@ -335,6 +354,34 @@ static void
 do_version(void)
 {
 	printf("carbon-c-relay v" VERSION " (" GIT_VERSION ")\n");
+#if defined(HAVE_GZIP) || defined(HAVE_LZ4) || defined(HAVE_SNAPPY) || defined(HAVE_SSL)
+	printf("enabled support for:");
+#endif
+#ifdef HAVE_GZIP
+	printf(" gzip");
+#endif
+#ifdef HAVE_LZ4
+	printf(" lz4");
+#endif
+#ifdef HAVE_SNAPPY
+	printf(" snappy");
+#endif
+#ifdef HAVE_SSL
+	printf(" ssl");
+#endif
+	printf("\n");
+
+	printf("regular expressions library:");
+#if defined(HAVE_ONIGURAMA)
+	printf(" oniguruma");
+#elif defined (HAVE_PCRE2)
+	printf(" PCRE2");
+#elif defined (HAVE_PCRE)
+	printf(" PCRE");
+#else
+	printf(" libc");
+#endif
+	printf("\n");
 
 	exit(0);
 }
@@ -354,12 +401,17 @@ do_usage(char *name, int exitcode)
 	printf("  -b  server send batch size, defaults to %d\n", batchsize);
 	printf("  -q  server queue size, defaults to %d\n", queuesize);
 	printf("  -L  server max stalls, defaults to %d\n", maxstalls);
-	printf("  -S  statistics sending interval in seconds, defaults to 60\n");
+#ifdef HAVE_SSL
+	printf("  -C  use CA <cert> to verify SSL connections\n");
+#endif
 	printf("  -B  connection listen backlog, defaults to 32\n");
 	printf("  -U  socket receive buffer size, max/min/default values depend on OS\n");
 	printf("  -T  IO timeout in milliseconds for server connections, defaults to %d\n", iotimeout);
-	printf("  -m  send statistics like carbon-cache.py, e.g. not cumulative\n");
+	printf("  -E  disable disconnecting idle connections after 10 minutes\n");
 	printf("  -c  characters to allow next to [A-Za-z0-9], defaults to -_:#\n");
+	printf("  -m  max string length of metric, defaults to 0, limit of -M\n");
+	printf("  -M  max string length of metric+ts+value+nl, defaults to %d\n",
+			METRIC_BUFSIZ);
 	printf("  -d  debug mode: currently writes statistics to log, prints hash\n"
 	       "      ring contents and matching position in test mode (-t)\n");
 #ifdef ENABLE_TRACE
@@ -369,7 +421,7 @@ do_usage(char *name, int exitcode)
 	       "      statistics, report drop counts and queue pressure to log\n");
 	printf("  -t  config test mode: prints rule matches from input on stdin\n");
 	printf("  -H  hostname: override hostname (used in statistics)\n");
-	printf("  -D  daemonise: run in a background\n");
+	printf("  -D  daemonise: detach and run in background\n");
 	printf("  -P  pidfile: write a pid to a specified pidfile\n");
 	printf("  -O  minimum number of rules before optimising the ruleset, default: %d\n", optimiserthreshold); 
 
@@ -389,11 +441,15 @@ main(int argc, char * const argv[])
 	int fds[2];
 	unsigned char startup_success = 0;
 	listener *lsnrs;
+	int maxmetriclen = -1;
+	int maxinplen = -1;
 
 	if (gethostname(relay_hostname, sizeof(relay_hostname)) < 0)
 		snprintf(relay_hostname, sizeof(relay_hostname), "127.0.0.1");
 
-	while ((ch = getopt(argc, argv, ":hvdstf:l:p:w:b:q:L:T:c:H:B:U:DP:O:")) != -1) {
+	while ((ch = getopt(argc, argv,
+					":hvdstf:l:p:w:b:q:L:C:T:c:m:M:H:B:U:EDP:O:")) != -1)
+	{
 		switch (ch) {
 			case 'v':
 				do_version();
@@ -455,12 +511,19 @@ main(int argc, char * const argv[])
 				break;
 			case 'L':
 				maxstalls = atoi(optarg);
-				if (maxstalls < 0 || maxstalls >= (1 << SERVER_STALL_BITS)) {
+				if (maxstalls < 0 || maxstalls >= (1 << SERVER_STALL_BITS) ||
+						!isdigit(*optarg))
+				{
 					fprintf(stderr, "error: maximum stalls needs to be a number "
 							"between 0 and %d\n", (1 << SERVER_STALL_BITS) - 1);
 					do_usage(argv[0], 1);
 				}
 				break;
+			case 'C':
+#ifdef HAVE_SSL
+				sslCA = optarg;
+#endif
+				break;
 			case 'T': {
 				int val = atoi(optarg);
 				if (val <= 0) {
@@ -474,9 +537,35 @@ main(int argc, char * const argv[])
 				}
 				iotimeout = (unsigned short)val;
 			}	break;
+			case 'E':
+				noexpire = 1;
+				break;
 			case 'c':
 				allowed_chars = optarg;
 				break;
+			case 'm': {
+				int val = atoi(optarg);
+				if (val < 0 || !isdigit(*optarg)) {
+					fprintf(stderr, "error: max metric length needs to "
+							"be a number >=0\n");
+					do_usage(argv[0], 1);
+				}
+				maxmetriclen = val;
+			}	break;
+			case 'M': {
+				int val = atoi(optarg);
+				if (val <= 0) {
+					fprintf(stderr, "error: max length needs to "
+							"be a number >0\n");
+					do_usage(argv[0], 1);
+				}
+				if (val > METRIC_BUFSIZ) {
+					fprintf(stderr, "error: max length cannot be >%d\n",
+							METRIC_BUFSIZ);
+					do_usage(argv[0], 1);
+				}
+				maxinplen = val;
+			}	break;
 			case 'H':
 				snprintf(relay_hostname, sizeof(relay_hostname), "%s", optarg);
 				break;
@@ -532,6 +621,11 @@ main(int argc, char * const argv[])
 				break;
 			case 'O': {
 				int val = atoi(optarg);
+				if (val == 0 && !isdigit(*optarg)) {
+					fprintf(stderr, "error: optimiser threshold needs to "
+							"be a number\n");
+					do_usage(argv[0], 1);
+				}
 				optimiserthreshold = val < 0 ? -1 : val;
 			}	break;
 			case '?':
@@ -555,7 +649,7 @@ main(int argc, char * const argv[])
 		workercnt = mode & MODE_SUBMISSION ? 2 : get_cores();
 
 	/* disable collector for submission mode */
-	if (mode & MODE_SUBMISSION)
+	if (mode & MODE_SUBMISSION && !(mode & MODE_DEBUG))
 		collector_interval = 0;
 
 	/* any_of failover maths need batchsize to be smaller than queuesize */
@@ -564,6 +658,18 @@ main(int argc, char * const argv[])
 		exit(1);
 	}
 
+	/* check bufsizes make sense */
+	if (maxinplen > 0 && maxmetriclen > 0 && maxmetriclen >= maxinplen)
+	{
+		fprintf(stderr, "error: max metric len must be smaller than "
+				"max length of input string\n");
+		exit(1);
+	}
+	if (maxinplen == -1)
+		maxinplen = METRIC_BUFSIZ;
+	if (maxmetriclen == -1)
+		maxmetriclen = maxinplen;  /* never hit the limit */
+
 	if (relay_logfile != NULL && (mode & MODE_TEST) == 0) {
 		FILE *f = fopen(relay_logfile, "a");
 		if (f == NULL) {
@@ -606,6 +712,20 @@ main(int argc, char * const argv[])
 		}
 	}
 
+#ifdef HAVE_SSL
+	/* check if we can read the given CA before starting up and stuff */
+	if (sslCA != NULL) {
+		struct stat st;
+		if (stat(sslCA, &st) == -1) {
+			fprintf(stderr, "failed to open TLS/SSL CA file '%s': %s\n",
+					sslCA, strerror(errno));
+			exit(1);
+		}
+		if (S_ISDIR(st.st_mode))
+			sslCAisdir = 1;
+	}
+#endif
+
 	if (mode & MODE_DAEMON) {
 		pid_t p;
 
@@ -719,11 +839,21 @@ main(int argc, char * const argv[])
 		fprintf(relay_stdout, "    listen backlog = %u\n", listenbacklog);
 		if (sockbufsize > 0)
 			fprintf(relay_stdout, "    socket bufsize = %u\n", sockbufsize);
+#ifdef HAVE_SSL
+		if (sslCA != NULL)
+			fprintf(relay_stdout, "    tls/ssl CA = %s\n", sslCA);
+#endif
 		fprintf(relay_stdout, "    server connection IO timeout = %dms\n",
 				iotimeout);
+		fprintf(relay_stdout, "    idle connections disconnect timeout = %s\n",
+				noexpire ? "never" : "10m");
 		if (allowed_chars != NULL)
 			fprintf(relay_stdout, "    extra allowed characters = %s\n",
 					allowed_chars);
+		if (maxinplen < METRIC_BUFSIZ)
+			fprintf(relay_stdout, "    max input length = %d\n", maxinplen);
+		if (maxmetriclen < maxinplen)
+			fprintf(relay_stdout, "    max metric length = %d\n", maxmetriclen);
 		if (mode & MODE_DEBUG)
 			fprintf(relay_stdout, "    debug = true\n");
 #ifdef ENABLE_TRACE
@@ -745,6 +875,11 @@ main(int argc, char * const argv[])
 	OpenSSL_add_all_algorithms();
 #endif
 
+#ifdef YYDEBUG
+	if (mode & MODE_TRACE)
+		router_yydebug = 1;
+#endif
+
 	if ((rtr = router_readconfig(NULL, config, workercnt,
 					queuesize, batchsize, maxstalls,
 					iotimeout, sockbufsize, listenport)) == NULL)
@@ -805,6 +940,9 @@ main(int argc, char * const argv[])
 	if (signal(SIGHUP, sig_handler) == SIG_ERR) {
 		exit_err("failed to create SIGHUP handler: %s\n", strerror(errno));
 	}
+	if (signal(SIGUSR1, sig_handler) == SIG_ERR) {
+		exit_err("failed to create SIGUSR1 handler: %s\n", strerror(errno));
+	}
 	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
 		exit_err("failed to ignore SIGPIPE: %s\n", strerror(errno));
 	}
@@ -838,7 +976,8 @@ main(int argc, char * const argv[])
 	logout("starting %d workers\n", workercnt);
 	for (id = 0; id < workercnt; id++) {
 		workers[id + 1] = dispatch_new_connection(
-				(unsigned char)id, rtr, allowed_chars);
+				(unsigned char)id, rtr, allowed_chars,
+				maxinplen, maxmetriclen);
 		if (workers[id + 1] == NULL) {
 			logerr("failed to add worker %d\n", id);
 			break;
@@ -904,7 +1043,6 @@ main(int argc, char * const argv[])
 	lsnrs = router_get_listeners(rtr);
 	for ( ; lsnrs != NULL; lsnrs = lsnrs->next) {
 		dispatch_removelistener(lsnrs);
-		shutdownclose(lsnrs);
 	}
 	/* since workers will be freed, stop querying the structures */
 	collector_stop();
