@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2017 Fabian Groffen
+ * Copyright 2013-2022 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -16,13 +16,13 @@
 
 
 #include <stdio.h>
+#include <stdint.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <string.h>
 #include <ctype.h>
 #include <glob.h>
-#include <regex.h>
 #include <sys/stat.h>
 #include <netdb.h>
 #include <sys/socket.h>
@@ -39,6 +39,7 @@
 #include "allocator.h"
 #include "relay.h"
 #include "router.h"
+#include "posixregex.h"
 #include "conffile.h"
 #include "conffile.tab.h"
 
@@ -81,6 +82,29 @@ int router_yylex_destroy(void *);
 /* catch string used for collector output */
 #define STUB_STATS  "_collector_stub_"
 
+/* string representations for symbols in the matching enums from
+ * router.h */
+const char *con_proto_str[] = {
+	/* 0 */ "<unset>",
+	/* 1 */ "tcp",
+	/* 2 */ "udp",
+	/* 3 */ "pipe",
+	/* 4 */ "file",
+	/* 5 */ "unix"
+};
+const char *con_type_str[] = {
+	/* 0 */ "<unset>",
+	/* 1 */ "linemode",
+	/* 2 */ "syslog"
+};
+const char *con_trnsp_str[] = {
+	/* 0 */ "<unset>",
+	/* 1 */ "plain",
+	/* 2 */ "gzip",
+	/* 3 */ "lz4",
+	/* 4 */ "snappy"
+};
+
 /**
  * Frees regexes from routes.
  */
@@ -295,7 +319,7 @@ router_yyerror(
 	regfree(&re);
 
 	/* clean up the keywords */
-	if (regcomp(&re, "cr([A-Z][A-Z0-9]*)", REG_EXTENDED) != 0) {
+	if (regcomp(&re, "cr([A-Z][A-Z0-9_]*)", REG_EXTENDED) != 0) {
 		r->parser_err.msg = ra_strdup(a, msg);
 		return;
 	}
@@ -372,10 +396,10 @@ router_yyerror(
 char *
 router_validate_address(
 		router *rtr,
-		char **retip, int *retport, void **retsaddr, void **rethint,
-		char *ip, con_proto proto)
+		char **retip, unsigned short *retport, void **retsaddr,
+		void **rethint, char *ip, con_proto proto)
 {
-	int port = GRAPHITE_PORT;
+	unsigned short port = GRAPHITE_PORT;
 	struct addrinfo *saddr = NULL;
 	struct addrinfo hint;
 	char sport[8];
@@ -389,11 +413,17 @@ router_validate_address(
 	if (*(p - 1) == ']')
 		lastcolon = NULL;
 	if (lastcolon != NULL) {
+		long lport;
 		char *endp = NULL;
 		*lastcolon = '\0';
-		port = (int)strtol(lastcolon + 1, &endp, 10);
-		if (port < 1 || endp != p)
-			return(ra_strdup(rtr->a, "invalid port number"));
+		lport = strtol(lastcolon + 1, &endp, 10);
+		if (lport < 1 || lport > UINT16_MAX || endp != p) {
+			char errbuf[40];
+			snprintf(errbuf, sizeof(errbuf),
+				 "invalid port number '%s'", lastcolon + 1);
+			return(ra_strdup(rtr->a, errbuf));
+		}
+		port = lport;
 	}
 	if (*ip == '[') {
 		ip++;
@@ -556,7 +586,7 @@ char *
 router_add_server(
 		router *ret,
 		char *ip,
-		int port,
+		unsigned short port,
 		char *inst,
 		con_type type,
 		con_trnsp transport,
@@ -602,7 +632,7 @@ router_add_server(
 			}
 		}
 		if (newserver == NULL) {
-			newserver = server_new(ip, (unsigned short)port,
+			newserver = server_new(ip, port,
 					type, transport, proto, walk, hint,
 					ret->conf.queuesize, ret->conf.batchsize,
 					ret->conf.maxstalls, ret->conf.iotimeout,
@@ -614,9 +644,7 @@ router_add_server(
 				snprintf(errbuf, sizeof(errbuf),
 						"failed to add server %s:%d (%s) "
 						"to cluster %s", ip, port,
-						proto == CON_UDP ? "udp" :
-						proto == CON_TCP ? "tcp" :
-						proto == CON_FILE ? "file" : "???", cl->name);
+						con_proto_str[proto], cl->name);
 				return ra_strdup(ret->a, errbuf);
 			}
 			if (ret->srvrs == NULL) {
@@ -630,7 +658,22 @@ router_add_server(
 			s->refcnt = 1;
 			s->server = newserver;
 		}
-		
+
+		/* check if the type matches */
+		if (s->refcnt > 1 && type != server_type(newserver)) {
+			freeaddrinfo(saddrs);
+			if (hint)
+				free(hint);
+			snprintf(errbuf, sizeof(errbuf),
+					"cannot set type '%s' for "
+					"server %s:%d: server was previously "
+					"defined with type '%s'",
+					con_type_str[type], serverip(newserver),
+					server_port(newserver),
+					con_type_str[server_type(newserver)]);
+			return ra_strdup(ret->a, errbuf);
+		}
+
 		/* check instance matches the existing server */
 		if (s->refcnt > 1) {
 			char *sinst = server_instance(newserver);
@@ -823,8 +866,11 @@ router_add_cluster(router *r, cluster *cl)
 		 * number of servers */
 		size_t i = 0;
 		char errbuf[512];
+		server **secondaries;
+
 		for (w = cl->members.ch->servers; w != NULL; w = w->next)
 			i++;
+
 		if (i < cl->members.ch->repl_factor) {
 			snprintf(errbuf, sizeof(errbuf),
 					"invalid cluster '%s': replication count (%u) is "
@@ -832,6 +878,16 @@ router_add_cluster(router *r, cluster *cl)
 					cl->name, cl->members.ch->repl_factor, i);
 			return ra_strdup(r->a, errbuf);
 		}
+
+		if (cl->isdynamic) {
+			secondaries = ra_malloc(r->a, sizeof(server *) * i);
+			if (secondaries == NULL)
+				return ra_strdup(r->a, "malloc failed for dynamic servers");
+			for (i = 0, w = cl->members.ch->servers; w != NULL; w = w->next)
+				secondaries[i++] = w->server;
+			for (w = cl->members.ch->servers; w != NULL; w = w->next)
+				server_add_secondaries(w->server, secondaries, i);
+		}
 	}
 	return NULL;
 }
@@ -958,9 +1014,13 @@ router_add_listener(
 		con_type ltype,
 		con_trnsp trnsp,
 		char *pemcert,
+		tlsprotover protomin,
+		tlsprotover protomax,
+		char *ciphers,
+		char *suites,
 		con_proto ctype,
 		char *ip,
-		int port,
+		unsigned short port,
 		struct addrinfo *saddrs)
 {
 	listener *lwalk;
@@ -1041,6 +1101,10 @@ router_add_listener(
 	lwalk->ctx = NULL;
 	lwalk->sslstrms = NULL;
 	lwalk->pemcert = pemcert;
+	lwalk->protomin = protomin;
+	lwalk->protomax = protomax;
+	lwalk->ciphers = ciphers;
+	lwalk->ciphersuites = suites;
 	if (stat(pemcert, &st) != -1) {
 		memcpy(&(lwalk->pemmtimespec), &(st.st_mtime),
 				sizeof(struct timespec));
@@ -1091,6 +1155,7 @@ router_readconfig(router *orig,
 {
 	FILE *cnf;
 	char *buf;
+	int fd;
 	size_t len = 0;
 	struct stat st;
 	router *ret = NULL;
@@ -1147,9 +1212,23 @@ router_readconfig(router *orig,
 		return ret;
 	} /* else: include path is a regular file path */
 
+	fd = open(path, O_RDONLY);
+	if (fd == -1) {
+		logerr("unable to open file '%s': %s\n", path, strerror(errno));
+		return NULL;
+	}
+
 	/* if there is no config, don't try anything */
-	if (stat(path, &st) == -1) {
+	if (fstat(fd, &st) == -1) {
 		logerr("unable to stat file '%s': %s\n", path, strerror(errno));
+		close(fd);
+		return NULL;
+	}
+
+	/* if we're not dealing with a file, say so */
+	if (!S_ISREG(st.st_mode)) {
+		logerr("cannot open non-regular file '%s'\n", path);
+		close(fd);
 		return NULL;
 	}
 
@@ -1160,11 +1239,13 @@ router_readconfig(router *orig,
 		/* get a return structure (and allocator) in place */
 		if ((ret = malloc(sizeof(router))) == NULL) {
 			logerr("malloc failed for router return struct\n");
+			close(fd);
 			return NULL;
 		}
 		ret->a = ra_new();
 		if (ret->a == NULL) {
 			logerr("malloc failed for allocator\n");
+			close(fd);
 			free(ret);
 			return NULL;
 		}
@@ -1175,8 +1256,9 @@ router_readconfig(router *orig,
 
 		err = router_set_collectorvals(ret, 60, "carbon.relays.\\.1", CUM);
 		if (err != NULL) {
-			logerr("setcollectorvals: %s\n");
+			logerr("setcollectorvals: %s\n", err);
 			router_free(ret);
+			close(fd);
 			return NULL;
 		}
 		ret->collector.stub = NULL;
@@ -1195,12 +1277,14 @@ router_readconfig(router *orig,
 		if (cl == NULL) {
 			logerr("malloc failed for blackhole cluster\n");
 			router_free(ret);
+			close(fd);
 			return NULL;
 		}
 		cl->name = ra_strdup(ret->a, "blackhole");
 		if (cl->name == NULL) {
 			logerr("malloc failed for blackhole cluster name\n");
 			router_free(ret);
+			close(fd);
 			return NULL;
 		}
 		cl->type = BLACKHOLE;
@@ -1216,6 +1300,7 @@ router_readconfig(router *orig,
 	if (palloc == NULL) {
 		logerr("malloc failed for parse allocator\n");
 		router_free(ret);
+		close(fd);
 		return NULL;
 	}
 
@@ -1223,13 +1308,15 @@ router_readconfig(router *orig,
 		logerr("malloc failed for config file buffer\n");
 		ra_free(palloc);
 		router_free(ret);
+		close(fd);
 		return NULL;
 	}
 
-	if ((cnf = fopen(path, "r")) == NULL) {
+	if ((cnf = fdopen(fd, "r")) == NULL) {
 		logerr("failed to open config file '%s': %s\n", path, strerror(errno));
 		ra_free(palloc);
 		router_free(ret);
+		close(fd);
 		return NULL;
 	}
 
@@ -1254,7 +1341,7 @@ router_readconfig(router *orig,
 			char *p;
 			char *carets;
 			size_t carlen;
-			fprintf(stderr, "%s:%zd:%zd: %s\n", path, ret->parser_err.line,
+			logerr("%s:%zd:%zd: %s\n", path, ret->parser_err.line,
 					ret->parser_err.start, ret->parser_err.msg);
 			/* get some relevant context from buff and put ^^^^ below it
 			 * to point out the position of the error */
@@ -1272,15 +1359,15 @@ router_readconfig(router *orig,
 			carets = ra_malloc(palloc, carlen + 1);
 			memset(carets, '^', carlen);
 			carets[carlen] = '\0';
-			fprintf(stderr, "%s\n", line);
+			logerr("%s\n", line);
 			/* deal with tabs in the input */
 			for (p = line; p - line < ret->parser_err.start; p++)
 				if (*p != '\t')
 					*p = ' ';
 			*p = '\0';
-			fprintf(stderr, "%s%s\n", line, carets);
+			logerr("%s%s\n", line, carets);
 		} else if (ret->parser_err.msg != NULL) {
-			fprintf(stderr, "%s: %s\n", path, ret->parser_err.msg);
+			logerr("%s: %s\n", path, ret->parser_err.msg);
 		}
 		/* clear the message, so it isn't printed when this was an
 		 * included config */
@@ -1298,7 +1385,7 @@ router_readconfig(router *orig,
 		if (ret->listeners == NULL) {
 			char sockbuf[128];
 			char *ip;
-			int port;
+			unsigned short port;
 			void *saddrs;
 			void *hint;
 
@@ -1307,21 +1394,21 @@ router_readconfig(router *orig,
 			router_validate_address(ret, &ip, &port, &saddrs, &hint,
 					sockbuf, CON_TCP);
 			free(hint);
-			router_add_listener(ret, T_LINEMODE, W_PLAIN, NULL, CON_TCP,
-					ip, port, saddrs);
+			router_add_listener(ret, T_LINEMODE, W_PLAIN, NULL, 0, 0,
+								NULL, NULL, CON_TCP, ip, port, saddrs);
 
 			hint = NULL;
 			snprintf(sockbuf, sizeof(sockbuf), ":%u", listenport);
 			router_validate_address(ret, &ip, &port, &saddrs, &hint,
 					sockbuf, CON_UDP);
 			free(hint);
-			router_add_listener(ret, T_LINEMODE, W_PLAIN, NULL, CON_UDP,
-					ip, port, saddrs);
+			router_add_listener(ret, T_LINEMODE, W_PLAIN, NULL, 0, 0,
+								NULL, NULL, CON_UDP, ip, port, saddrs);
 
 			snprintf(sockbuf, sizeof(sockbuf), "%s/%s.%u",
 					TMPDIR, SOCKFILE, listenport);
-			router_add_listener(ret, T_LINEMODE, W_PLAIN, NULL, CON_UNIX,
-					sockbuf, 0, NULL);
+			router_add_listener(ret, T_LINEMODE, W_PLAIN, NULL, 0, 0,
+								NULL, NULL, CON_UNIX, sockbuf, 0, NULL);
 		}
 	}
 
@@ -1544,9 +1631,10 @@ router_optimise(router *r, int threshold)
 			rnext = rwalk->next;
 			rwalk->next = NULL;
 			if (rwalk->stop) {
-				bstart = blast;
+				bstart = NULL;
 				rlast = NULL;
 				seq++;
+				tracef("terminating group due to stop rule\n");
 			}
 			continue;
 		}
@@ -1565,6 +1653,9 @@ router_optimise(router *r, int threshold)
 					bwalk->next = NULL;
 				}
 				rlast = rwalk;
+				bstart = NULL;
+				seq++;  /* finish this group */
+				tracef("finishing group due to stop rule\n");
 			}
 		} else {
 			bstart = blast;
@@ -1579,14 +1670,21 @@ router_optimise(router *r, int threshold)
 	blast->next = NULL;
 	blast->seqnr = seq;
 
+	tracef("reducing into groups\n");
 	rwalk = r->routes = NULL;
 	seq = 1;
 	/* create groups, if feasible */
 	for (bwalk = blocks; bwalk != NULL; bwalk = blast) {
-		if (bwalk->seqnr != seq) {
+		tracef("seq=%zd, bwalk=%p, ->seqnr=%zd, ->refcnt=%zd, ->pattern=%s\n",
+				seq, bwalk, bwalk->seqnr, bwalk->refcnt, bwalk->pattern);
+		if (bwalk->seqnr > seq) {
 			seq++;
 			blast = bwalk;
 			continue;
+		} else if (bwalk->seqnr < seq) {
+			logerr("routes optimisation lost track, "
+					"skipping further optimisations\n");
+			return;
 		}
 
 		if (bwalk->refcnt == 0) {
@@ -1700,7 +1798,7 @@ router_getcollectormode(router *rtr)
  * Sets the collector interval in seconds, the metric prefix string, and
  * the emission mode (cumulative or sum).
  */
-inline char *
+char *
 router_set_collectorvals(router *rtr, int intv, char *prefix, col_mode smode)
 {
 	if (intv >= 0)
@@ -1712,12 +1810,20 @@ router_set_collectorvals(router *rtr, int intv, char *prefix, col_mode smode)
 		char *dummy = relay_hostname + strlen(relay_hostname);
 		size_t nmatch = 3;
 		regmatch_t pmatch[3];
+		int reret;
 
-		if (regcomp(&re, expr, REG_EXTENDED) != 0)
-			return ra_strdup(rtr->a, "failed to compile hostname regexp");
-		if (regexec(&re, relay_hostname, nmatch, pmatch, 0) != 0) {
+		if ((reret = regcomp(&re, expr, REG_EXTENDED)) != 0) {
+			size_t len = snprintf(cprefix, sizeof(cprefix),
+					"failed to compile hostname regexp: ");
+			regerror(reret, &re, cprefix + len, sizeof(cprefix) - len);
+			return ra_strdup(rtr->a, cprefix);
+		}
+		if ((reret = regexec(&re, relay_hostname, nmatch, pmatch, 0)) != 0) {
+			size_t len = snprintf(cprefix, sizeof(cprefix),
+					"failed to execute hostname regexp: ");
+			regerror(reret, &re, cprefix + len, sizeof(cprefix) - len);
 			regfree(&re);
-			return ra_strdup(rtr->a, "failed to execute hostname regexp");
+			return ra_strdup(rtr->a, cprefix);
 		}
 		if (router_rewrite_metric(
 				&cprefix, &dummy,
@@ -1737,6 +1843,74 @@ router_set_collectorvals(router *rtr, int intv, char *prefix, col_mode smode)
 	return NULL;
 }
 
+static char *_router_quoteident_buf = NULL;
+static size_t _router_quoteident_buflen = 0;
+static const char *router_quoteident(const char *ident)
+{
+	/*
+	IDCHR   [a-zA-Z0-9[\](){}|\\^?*$&%<>:/,._+-]
+	ident = {IDCHR}+(\\[ ]{IDCHR}+)*
+	*/
+	char allowodd[] = { '[', ']', '(', ')', '{', '}',
+		'|', '\\', '^', '?', '*', '$', '%', '<', '>',
+		':', '/', ',', '.', '_', '+', '-', '\0' /* string terminator */};
+	size_t idlen = strlen(ident);
+	size_t idpos;
+	size_t bufpos;
+	char needquote = 0;
+
+	/* parser won't accept empty strings for input, so we can assume
+	 * idlen >= 1 here */
+
+	/* setup buf for worst case, allocate in 32-byte blocks to try and
+	 * avoid reallocs for just a few bytes */
+	if (_router_quoteident_buflen < 1 + (idlen * 2) + 1 + 1) {
+		_router_quoteident_buflen = (((1 + (idlen * 2) + 1 + 1) / 32) + 1) * 32;
+		_router_quoteident_buf =
+			realloc(_router_quoteident_buf, _router_quoteident_buflen);
+	}
+
+	bufpos = 0;
+	_router_quoteident_buf[bufpos++] = '"';
+
+	for (idpos = 0; idpos < idlen; idpos++) {
+		if (!(
+				(ident[idpos] >= 'a' && ident[idpos] <= 'z') ||
+				(ident[idpos] >= 'A' && ident[idpos] <= 'Z') ||
+				(ident[idpos] >= '0' && ident[idpos] <= '9') ||
+				strchr(allowodd, (int)ident[idpos]) != NULL)
+		   )
+		{
+			if (!(idpos > 1 && ident[idpos] == ' ' && ident[idpos - 1] == '\\'))
+				needquote = 1;
+		}
+		_router_quoteident_buf[bufpos++] = ident[idpos];
+	}
+	_router_quoteident_buf[bufpos++] = '"';
+	_router_quoteident_buf[bufpos] = '\0';
+
+	if (needquote)
+		return _router_quoteident_buf;
+
+	return ident;
+}
+
+#ifdef HAVE_SSL
+const char *
+router_tlsprotostr(tlsprotover protover)
+{
+	switch (protover) {
+		case _rp_UNSET:   return "";         break;
+		case _rp_SSL3:    return "ssl3";     break;
+		case _rp_TLS1_0:  return "tls1.0";   break;
+		case _rp_TLS1_1:  return "tls1.1";   break;
+		case _rp_TLS1_2:  return "tls1.2";   break;
+		case _rp_TLS1_3:  return "tls1.3";   break;
+		default:          return "unknown";  break;
+	}
+}
+#endif
+
 /**
  * Mere debugging function to check if the configuration is picked up
  * alright.  If all is set to false, aggregation rules won't be printed.
@@ -1753,49 +1927,60 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 	/* start with configuration wise standard components */
 #define PPROTO \
 	server_ctype(s->server) == CON_UDP ? " proto udp" : ""
+#define PTYPEFMT "%s%s"
 #define PTYPE \
-	server_type(s->server) == T_LINEMODE ? "" : " type ???"
+	server_type(s->server) == T_LINEMODE ? "" : " type ", \
+	server_type(s->server) == T_LINEMODE ? "" : \
+		con_type_str[server_type(s->server)]
+#define PTRNSPFMT "%s%s%s"
 #define PTRNSP \
-	server_transport(s->server) == W_PLAIN ? "" : \
-	server_transport(s->server) == W_GZIP  ? " transport gzip" : \
-	server_transport(s->server) == W_LZ4   ? " transport lz4" : \
-	server_transport(s->server) == W_SSL   ? " transport ssl" : " unknown"
+	server_transport(s->server) == W_PLAIN ? "" : " transport", \
+	(server_transport(s->server) & 0xFFFF) == W_PLAIN ? "" : \
+		con_trnsp_str[server_transport(s->server) & 0xFFFF], \
+	(server_transport(s->server) & ~0xFFFF) == W_SSL ? " ssl" : ""
 
 	if (rtr->listeners != NULL) {
 		listener *walk;
 		fprintf(f, "listen\n");
 		for (walk = rtr->listeners; walk != NULL; walk = walk->next) {
 			if (walk->lsnrtype == T_LINEMODE) {
-				fprintf(f, "    type linemode%s%s%s",
-						walk->transport == W_PLAIN ? "" :
-						walk->transport == W_GZIP  ? " transport gzip" :
-						walk->transport == W_LZ4   ? " transport lz4" :
-						walk->transport == W_SSL   ? " transport ssl" :
-						                             " unknown",
-						walk->transport == W_SSL ? " " : "",
+				fprintf(f, "    type linemode");
+				if ((walk->transport & 0xFFFF) != W_PLAIN)
+					fprintf(f, " transport %s",
+							con_trnsp_str[walk->transport & 0xFFFF]);
 #ifdef HAVE_SSL
-						walk->transport == W_SSL ? walk->pemcert : ""
-#else
-						""
-#endif
-						);
-#ifdef HAVE_SSL
-				if (walk->transport == W_SSL && pmode & PMODE_PEMT) {
-					fprintf(f, "  # mtime %ld.%ld",
-							walk->pemmtimespec.tv_sec,
-							walk->pemmtimespec.tv_nsec);
+				if ((walk->transport & ~0xFFFF) == W_SSL) {
+					fprintf(f, " ssl %s", router_quoteident(walk->pemcert));
+					if (walk->protomin != _rp_UNSET)
+						fprintf(f, " protomin %s",
+								router_tlsprotostr(walk->protomin));
+					if (walk->protomax != _rp_UNSET)
+						fprintf(f, " protomax %s",
+								router_tlsprotostr(walk->protomax));
+					if (walk->ciphers != NULL)
+						fprintf(f, " ciphers %s",
+								router_quoteident(walk->ciphers));
+					if (walk->ciphersuites != NULL)
+						fprintf(f, " ciphersuites %s",
+								router_quoteident(walk->ciphersuites));
+					if (pmode & PMODE_PEMT)
+					{
+						fprintf(f, "  # mtime %ld.%ld",
+								walk->pemmtimespec.tv_sec,
+								walk->pemmtimespec.tv_nsec);
+					}
 				}
 #endif
 				fprintf(f, "\n");
 				do {
 					if (walk->ctype == CON_UNIX) {
-						fprintf(f, "        %s proto unix\n", walk->ip);
+						fprintf(f, "        %s proto unix\n",
+								router_quoteident(walk->ip));
 					} else {
 						fprintf(f, "        %s%s%u proto %s\n",
 								walk->ip == NULL ? "" : walk->ip,
 								walk->ip == NULL ? "" : ":",
-								walk->port,
-								walk->ctype == CON_UDP ? "udp" : "tcp");
+								walk->port, con_proto_str[walk->ctype]);
 					}
 				} while (walk->next != NULL
 						&& walk->lsnrtype == walk->next->lsnrtype
@@ -1810,17 +1995,19 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 				rtr->collector.interval);
 		if (rtr->collector.mode == SUB)
 			fprintf(f, "    reset counters after interval\n");
-		fprintf(f, "    prefix with %s\n", rtr->collector.prefix);
+		fprintf(f, "    prefix with %s\n",
+				router_quoteident(rtr->collector.prefix));
 		if (rtr->collector.stub != NULL) {
 			fprintf(f, "    send to");
 			for (r = rtr->routes; r != NULL; r = r->next) {
 				if (r->dests->cl->type == STATSTUB) {
 					destinations *d = r->dests->cl->members.routes->dests;
 					if (d->next == NULL) {
-						fprintf(f, " %s", d->cl->name);
+						fprintf(f, " %s", router_quoteident(d->cl->name));
 					} else {
 						for ( ; d != NULL; d = d->next)
-							fprintf(f, "\n        %s", d->cl->name);
+							fprintf(f, "\n        %s",
+									router_quoteident(d->cl->name));
 					}
 					fprintf(f, "\n    stop\n");
 					break;
@@ -1837,37 +2024,39 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 				c->type == GROUP || c->type == AGGREGATION ||
 				c->type == AGGRSTUB || c->type == STATSTUB)
 			continue;
-		fprintf(f, "cluster %s\n", c->name);
+		fprintf(f, "cluster %s\n", router_quoteident(c->name));
 		if (c->type == FORWARD) {
 			fprintf(f, "    forward\n");
 			for (s = c->members.forward; s != NULL; s = s->next)
-				fprintf(f, "        %s:%d%s%s%s\n",
+				fprintf(f, "        %s:%d%s" PTYPEFMT PTRNSPFMT "\n",
 						serverip(s->server), server_port(s->server),
 						PPROTO, PTYPE, PTRNSP);
 		} else if (c->type == FILELOG || c->type == FILELOGIP) {
 			fprintf(f, "    file%s\n", c->type == FILELOGIP ? " ip" : "");
 			for (s = c->members.forward; s != NULL; s = s->next)
 				fprintf(f, "        %s\n",
-						serverip(s->server));
+						router_quoteident(serverip(s->server)));
 		} else if (c->type == ANYOF || c->type == FAILOVER) {
 			fprintf(f, "    %s\n", c->type == ANYOF ? "any_of" : "failover");
 			for (s = c->members.anyof->list; s != NULL; s = s->next)
-				fprintf(f, "        %s:%d%s%s%s\n",
+				fprintf(f, "        %s:%d%s" PTYPEFMT PTRNSPFMT "\n",
 						serverip(s->server), server_port(s->server),
 						PPROTO, PTYPE, PTRNSP);
 		} else if (c->type == CARBON_CH ||
 				c->type == FNV1A_CH ||
 				c->type == JUMP_CH)
 		{
-			fprintf(f, "    %s_ch replication %d\n",
+			fprintf(f, "    %s_ch replication %d%s\n",
 					c->type == CARBON_CH ? "carbon" : 
 					c->type == FNV1A_CH ? "fnv1a" : "jump_fnv1a",
-					c->members.ch->repl_factor);
+					c->members.ch->repl_factor,
+					c->isdynamic ? " dynamic" : "");
 			for (s = c->members.ch->servers; s != NULL; s = s->next)
-				fprintf(f, "        %s:%d%s%s%s%s%s\n",
+				fprintf(f, "        %s:%d%s%s%s" PTYPEFMT PTRNSPFMT "\n",
 						serverip(s->server), server_port(s->server),
 						server_instance(s->server) ? "=" : "",
-						server_instance(s->server) ? server_instance(s->server) : "",
+						server_instance(s->server) ?
+							server_instance(s->server) : "",
 						PPROTO, PTYPE, PTRNSP);
 		}
 		fprintf(f, "    ;\n");
@@ -1901,11 +2090,11 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 
 			fprintf(f, "aggregate");
 			if (r->next == NULL || r->next->dests->cl != aggr) {
-				fprintf(f, " %s\n", r->pattern);
+				fprintf(f, " %s\n", router_quoteident(r->pattern));
 			} else {
 				fprintf(f, "\n");
 				do {
-					fprintf(f, "        %s\n", r->pattern);
+					fprintf(f, "        %s\n", router_quoteident(r->pattern));
 				} while (r->next != NULL && r->next->dests->cl == aggr
 						&& (r = r->next) != NULL);
 			}
@@ -1918,7 +2107,8 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 					aggr->members.aggregation->tswhen == TS_MIDDLE ? "middle" :
 					aggr->members.aggregation->tswhen == TS_END ? "end" :
 					"<unknown>");
-			for (ac = aggr->members.aggregation->computes; ac != NULL; ac = ac->next) {
+			ac = aggr->members.aggregation->computes;
+			for ( ; ac != NULL; ac = ac->next) {
 				snprintf(percentile, sizeof(percentile),
 						"percentile%d", ac->percentile);
 				fprintf(f, "    compute %s write to\n"
@@ -1931,23 +2121,25 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 						ac->type == VAR ? "variance" :
 						ac->type == SDEV ? "stddev" :
 						"<unknown>",
-						ac->metric + strlen(stubname));
+						router_quoteident(ac->metric + strlen(stubname)));
 			}
 			if (!(pmode & PMODE_STUB) && r->dests->next != NULL) {
 				destinations *dn = r->dests->next;
 				fprintf(f, "    send to");
 				if (dn->next == NULL) {
-					fprintf(f, " %s\n", dn->cl->name);
+					fprintf(f, " %s\n", router_quoteident(dn->cl->name));
 				} else {
 					for (; dn != NULL; dn = dn->next)
-						fprintf(f, "\n        %s", dn->cl->name);
+						fprintf(f, "\n        %s",
+								router_quoteident(dn->cl->name));
 					fprintf(f, "\n");
 				}
 			}
 			fprintf(f, "%s    ;\n", r->stop ? "    stop\n" : "");
 		} else if (r->dests->cl->type == REWRITE) {
-			fprintf(f, "rewrite %s\n    into %s\n    ;\n",
-					r->pattern, r->dests->cl->members.replacement);
+			fprintf(f, "rewrite %s\n", router_quoteident(r->pattern));
+			fprintf(f, "    into %s\n    ;\n",
+					router_quoteident(r->dests->cl->members.replacement));
 		} else if (r->dests->cl->type == GROUP) {
 			size_t cnt = 0;
 			route *rwalk;
@@ -1955,7 +2147,8 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 			char *b = &blockname[sizeof(blockname) - 1];
 			char *p;
 
-			for (rwalk = r->dests->cl->members.routes; rwalk != NULL; rwalk = rwalk->next)
+			rwalk = r->dests->cl->members.routes;
+			for ( ; rwalk != NULL; rwalk = rwalk->next)
 				cnt++;
 			/* reverse the name, to make it human consumable */
 			*b-- ='\0';
@@ -1978,14 +2171,16 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 			if (pmode & PMODE_STUB) {
 				fprintf(f, "# stub match for aggregate/statistics rule "
 						"with send to\n");
-				fprintf(f, "match ^%s\n    send to", r->pattern);
+				fprintf(f, "match ^%s\n    send to",
+						router_quoteident(r->pattern));
 				if (r->dests->cl->members.routes->dests->next == NULL) {
-					fprintf(f, " %s",
-							r->dests->cl->members.routes->dests->cl->name);
+					fprintf(f, " %s", router_quoteident(
+							r->dests->cl->members.routes->dests->cl->name));
 				} else {
 					destinations *d = r->dests->cl->members.routes->dests;
 					for (; d != NULL; d = d->next)
-						fprintf(f, "\n        %s", d->cl->name);
+						fprintf(f, "\n        %s",
+								router_quoteident(d->cl->name));
 				}
 				fprintf(f, "%s\n    ;\n", r->stop ? "\n    stop" : "");
 			}
@@ -1995,12 +2190,14 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 			fprintf(f, "match");
 			if (r->next == NULL || r->next->dests != or->dests) {
 				fprintf(f, " %s\n",
-						r->matchtype == MATCHALL ? "*" : r->pattern);
+						r->matchtype == MATCHALL ? "*" :
+						router_quoteident(r->pattern));
 			} else {
 				fprintf(f, "\n");
 				do {
 					fprintf(f, "        %s\n",
-							r->matchtype == MATCHALL ? "*" : r->pattern);
+							r->matchtype == MATCHALL ? "*" :
+							router_quoteident(r->pattern));
 				} while (r->next != NULL && r->next->dests == or->dests
 						&& (r = r->next) != NULL);
 			}
@@ -2008,20 +2205,21 @@ router_printconfig(router *rtr, FILE *f, char pmode)
 			if (d->cl->type == VALIDATION) {
 				validate *v = d->cl->members.validation;
 				fprintf(f, "    validate %s else %s\n",
-						v->rule->pattern,
+						router_quoteident(v->rule->pattern),
 						v->action == VAL_LOG ? "log" : "drop");
 				/* hide this pseudo target */
 				d = d->next;
 			}
 			if (r->masq != NULL)
-				fprintf(f, "    route using %s\n", r->masq);
+				fprintf(f, "    route using %s\n", router_quoteident(r->masq));
 			if (d != NULL) {
 				fprintf(f, "    send to");
 				if (d->next == NULL) {
-					fprintf(f, " %s", d->cl->name);
+					fprintf(f, " %s", router_quoteident(d->cl->name));
 				} else {
 					for ( ; d != NULL; d = d->next)
-						fprintf(f, "\n        %s", d->cl->name);
+						fprintf(f, "\n        %s",
+								router_quoteident(d->cl->name));
 				}
 				fprintf(f, "\n%s    ;\n", or->stop ? "    stop\n" : "");
 			} else {
@@ -2045,7 +2243,7 @@ router_printdiffs(router *old, router *new, FILE *out)
 	char *tmp = getenv("TMPDIR");
 	char patho[512];
 	char pathn[512];
-	char buf[1024];
+	char buf[1033];
 	size_t len;
 	int ret;
 	mode_t mask;
@@ -2094,6 +2292,10 @@ router_printdiffs(router *old, router *new, FILE *out)
 	/* diff and print its output */
 	snprintf(buf, sizeof(buf), "diff -u %s %s", patho, pathn);
 	f = popen(buf, "r");
+	if (f == NULL) {
+		logerr("failed to popen diff: %s\n", strerror(errno));
+		return 1;
+	}
 	while ((len = fread(buf, 1, sizeof(buf) - 1, f)) > 0) {
 		if (fwrite(buf, len, 1, out) == 0)
 			/* ignore */
@@ -2207,11 +2409,14 @@ router_contains_listener(router *rtr, listener *lsnr)
 				if (lsnr->port == rwalk->port && strcmp(l, r) == 0) {
 #ifdef HAVE_SSL
 					/* check pemmtimespec */
-					if (lsnr->transport == W_SSL &&
-							lsnr->pemmtimespec.tv_sec ==
-									rwalk->pemmtimespec.tv_sec &&
-							lsnr->pemmtimespec.tv_nsec ==
-									rwalk->pemmtimespec.tv_nsec)
+					if ((lsnr->transport & ~0xFFFF) == W_SSL &&
+							(lsnr->pemmtimespec.tv_sec !=
+									rwalk->pemmtimespec.tv_sec || 
+							lsnr->pemmtimespec.tv_nsec !=
+									rwalk->pemmtimespec.tv_nsec))
+					{
+						continue;
+					} else
 #endif
 					{
 						match = 1;
@@ -2359,7 +2564,7 @@ router_metric_matches(
 	return ret;
 }
 
-inline size_t
+size_t
 router_rewrite_metric(
 		char (*newmetric)[METRIC_BUFSIZ],
 		char **newfirstspace,
@@ -2512,6 +2717,60 @@ router_rewrite_metric(
 	return 0;  /* we couldn't copy everything */
 }
 
+static inline const char *
+router_format_linemode(
+		const char *in,
+		char *out,
+		size_t *outlen,
+		char *srcaddr)
+{
+	(void)out;
+	*outlen = strlen(in);
+	return in;
+}
+
+static inline const char *
+router_format_linemodeip(
+		const char *in,
+		char *out,
+		size_t *outlen,
+		char *srcaddr)
+{
+	*outlen = snprintf(out, *outlen, "%s %s", srcaddr, in);
+	return out;
+}
+
+static inline const char *
+router_format_syslog(
+		const char *in,
+		char *out,
+		size_t *outlen,
+		char *srcaddr)
+{
+	time_t now;
+	struct tm tm_now;
+	char ts_now[32];
+
+	time(&now);
+	gmtime_r(&now, &tm_now);
+	/* https://tools.ietf.org/html/rfc5424#section-6.2.3 */
+	strftime(ts_now, sizeof(ts_now), "%Y-%m-%dT%H:%M:%S.000Z", &tm_now);
+	/* https://tools.ietf.org/html/rfc5424#section-6
+	 * PRI = 3 (system daemons) * 8 + 6 (informational)
+	 * VERSION = 1
+	 * TIMESTAMP = see above
+	 * HOSTNAME = src address
+	 * APP-NAME = carbon-c-relay
+	 * PROCID = - (NILVALUE)
+	 * MSGID = - (NILVALUE)
+	 * STRUCTURED-DATA = - (NILVALUE)
+	 * MSG = metric */
+	*outlen = snprintf(out, *outlen,
+			"<30>1 %s %s carbon-c-relay - - - %s",
+			ts_now, srcaddr, in);
+	return out;
+}
+
 static char
 router_route_intern(
 		char *blackholed,
@@ -2533,6 +2792,7 @@ router_route_intern(
 	const char *t;
 	char newmetric[METRIC_BUFSIZ];
 	char *newfirstspace = NULL;
+	const char *fmtmetric;
 	size_t len;
 	regmatch_t pmatch[RE_MAX_MATCHES];
 
@@ -2543,7 +2803,6 @@ router_route_intern(
 		return 1; \
 	}
 
-
 	for (w = r; w != NULL; w = w->next) {
 		if (w->dests->cl->type == GROUP) {
 			/* strrstr doesn't exist, grrr
@@ -2575,23 +2834,41 @@ router_route_intern(
 						firstspace,
 						w->dests->cl->members.routes,
 						dispatcher_id);
-		} else if (router_metric_matches(w, metric, firstspace, pmatch, dispatcher_id)) {
+		} else if (router_metric_matches(w, metric, firstspace,
+					pmatch, dispatcher_id))
+		{
 			stop = w->stop;
 			/* rule matches, send to destination(s) */
 			for (d = w->dests; d != NULL; d = d->next) {
+#define produce_metric(RET) \
+				len = sizeof(newmetric); \
+				if (server_type(RET.dest) == T_SYSLOGMODE) { \
+					fmtmetric = router_format_syslog( \
+							metric, newmetric, &len, srcaddr); \
+				} else { \
+					fmtmetric = router_format_linemode( \
+							metric, newmetric, &len, srcaddr); \
+				}
+#define set_metric(RET) \
+				RET.metric = malloc(sizeof(char) * len + sizeof(len)); \
+				*((size_t *)RET.metric) = len; \
+				memcpy((void *)(RET.metric + sizeof(len)), fmtmetric, len);
+
 				switch (d->cl->type) {
 					case BLACKHOLE:
 						*blackholed = 1;
 						break;
 					case FILELOGIP: {
 						servers *s;
-						snprintf(newmetric, sizeof(newmetric), "%s %s",
-								srcaddr, metric);
+						len = sizeof(newmetric);
+						fmtmetric = router_format_linemodeip(
+								metric, newmetric, &len, srcaddr);
 						for (s = d->cl->members.forward; s != NULL; s = s->next)
 						{
 							failif(retsize, *curlen + 1);
 							ret[*curlen].dest = s->server;
-							ret[(*curlen)++].metric = strdup(newmetric);
+							set_metric(ret[*curlen]);
+							(*curlen)++;
 						}
 						wassent = 1;
 					}	break;
@@ -2603,7 +2880,9 @@ router_route_intern(
 						{
 							failif(retsize, *curlen + 1);
 							ret[*curlen].dest = s->server;
-							ret[(*curlen)++].metric = strdup(metric);
+							produce_metric(ret[*curlen]);
+							set_metric(ret[*curlen]);
+							(*curlen)++;
 						}
 						wassent = 1;
 					}	break;
@@ -2623,7 +2902,9 @@ router_route_intern(
 						failif(retsize, *curlen + 1);
 						ret[*curlen].dest =
 							d->cl->members.anyof->servers[hash];
-						ret[(*curlen)++].metric = strdup(metric);
+						produce_metric(ret[*curlen]);
+						set_metric(ret[*curlen]);
+						(*curlen)++;
 						wassent = 1;
 					}	break;
 					case FAILOVER: {
@@ -2643,7 +2924,10 @@ router_route_intern(
 							/* all failed, take first server */
 							ret[*curlen].dest =
 								d->cl->members.anyof->servers[0];
-						ret[(*curlen)++].metric = strdup(metric);
+
+						produce_metric(ret[*curlen]);
+						set_metric(ret[*curlen]);
+						(*curlen)++;
 						wassent = 1;
 					}	break;
 					case CARBON_CH:
@@ -2673,8 +2957,11 @@ router_route_intern(
 								d->cl->members.ch->repl_factor,
 								w->masq ? newmetric : metric,
 								w->masq ? newfirstspace : firstspace);
-						for (i = 0; i < d->cl->members.ch->repl_factor; i++)
-							ret[(*curlen)++].metric = strdup(metric);
+						for (i = 0; i < d->cl->members.ch->repl_factor; i++) {
+							produce_metric(ret[*curlen]);
+							set_metric(ret[*curlen]);
+							(*curlen)++;
+						}
 						wassent = 1;
 					}	break;
 					case AGGREGATION: {
@@ -2724,6 +3011,10 @@ router_route_intern(
 								firstspace,
 								w->dests->cl->members.routes,
 								dispatcher_id);
+						/* ensure recursion doesn't result in false
+						 * blachole stats #405 */
+						if (*blackholed == 0)
+							wassent = 1;
 					}	break;
 					case VALIDATION: {
 						/* test whether data matches, if not, either log
@@ -2887,8 +3178,8 @@ router_test_intern(char *metric, char *firstspace, route *routes)
 						if (mode & MODE_DEBUG || d->next == NULL) {
 							stublen = 0;
 						} else {
-							char x;
-							stublen = snprintf(&x, 1,
+							char x[32];
+							stublen = snprintf(x, 32,
 									STUB_AGGR "%p__",
 									d->cl->members.aggregation);
 						}
