@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2017 Fabian Groffen
+ * Copyright 2013-2018 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -19,12 +19,12 @@
 #define ROUTER_H 1
 
 #include <stdio.h>
-#include <regex.h>
 #include <time.h>
 
 #include "relay.h"
 #include "server.h"
 #include "aggregator.h"
+#include "posixregex.h"
 
 #ifdef HAVE_SSL
 #include <openssl/ssl.h>
@@ -50,17 +50,44 @@ typedef struct {
 	server *dest;
 } destination;
 
+/* ensure we always define the protovers such that the parser can
+ * reference them, even though they will never get used */
+#ifdef HAVE_SSL
+# define TLSVER(S,V) S = V
+# ifndef SSL3_VERSION
+#  define SSL3_VERSION   0x0300
+# endif
+# ifndef TLS1_3_VERSION
+#  define TLS1_3_VERSION 0x0304
+# endif
+#else
+# define TLSVER(S,V) S
+#endif
+typedef enum _router_tlsprotover {
+	       _rp_UNSET = 0,
+	TLSVER(_rp_SSL3,   SSL3_VERSION),
+	TLSVER(_rp_TLS1_0, TLS1_VERSION),
+	TLSVER(_rp_TLS1_1, TLS1_1_VERSION),
+	TLSVER(_rp_TLS1_2, TLS1_2_VERSION),
+	TLSVER(_rp_TLS1_3, TLS1_3_VERSION)
+} tlsprotover;
+#undef TLSVER
+
 typedef struct _router_listener {
 	con_type lsnrtype;
 	con_trnsp transport;
 	con_proto ctype;
 	char *ip;
-	int port;
+	unsigned short port;
 	int *socks;
 #ifdef HAVE_SSL
 	SSL_CTX *ctx;
 	SSL **sslstrms;
 	char *pemcert;
+	tlsprotover protomin;
+	tlsprotover protomax;
+	char *ciphers;
+	char *ciphersuites;
 	struct timespec pemmtimespec;
 #endif
 	struct addrinfo *saddrs;
